# Copyright 2024 Albert Nieto

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import logging

logger = logging.getLogger(__name__)


def get_gpu_info():
    try:
        gpus = jax.devices("gpu")
        num_gpus = len(gpus)
        gpu_names = [gpu.device_kind for gpu in gpus]
        return num_gpus, gpu_names
    except Exception as e:
        logger.error(f"Error retrieving GPU information: {e}")
        return 0, []
