# Copyright 2024 Albert Nieto

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from IPython.core.magic import register_line_magic
from IPython.display import display, HTML


@register_line_magic
def scroll_output(line):
    """Magic command to enable scrolling in output cells with a specified max height."""
    height = line.strip() or "300px"
    display(
        HTML(
            f"""
        <style>
            .output_area pre {{
                max-height: {height};
                overflow-y: scroll;
            }}
            .output_area {{
                max-height: {height};
                overflow-y: scroll;
            }}
        </style>
    """
        )
    )


# Now you can use the magic command in your notebook
